/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.news;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.io.IOException;
import java.net.URL;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.util.thread.ThreadType;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.semantic.document.Document;
import it.tidalwave.semantic.rss.RssFeed;
import it.tidalwave.bluebill.mobile.preferences.NetworkingPreferences;
import it.tidalwave.bluebill.mobile.news.impl.Cache;
import it.tidalwave.bluebill.mobile.news.impl.ReadMessagesLog;
import static it.tidalwave.bluebill.mobile.news.impl.Cache.Status.*;

/***********************************************************************************************************************
 *
 * 
 * @stereotype Service
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@ServiceProvider(service=NewsService.class)
public class DefaultNewsService implements NewsService
  {
    private static final Logger log = LoggerFactory.getLogger(DefaultNewsService.class);

    private static final Class<DefaultNewsService> _ = DefaultNewsService.class;

    private static final String NEWS_FEED_URL = "http://bluebill.tidalwave.it/mobile/blog.rss";

    @CheckForNull
    protected Thread startThread;

    @Nonnull
    private final Provider<NetworkingPreferences> networkingPreferences = Locator.createProviderFor(NetworkingPreferences.class);

    /* package */ final Cache cache; 
    
    /* package */  final ReadMessagesLog readMessagesLog;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ThreadType.ANY)
    public DefaultNewsService()
      throws IOException
      {
        this.cache = new Cache(createUrl(NEWS_FEED_URL));
        readMessagesLog = new ReadMessagesLog();
      }

    /*******************************************************************************************************************
     *
     * For tests only.
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ThreadType.ANY)
    /* package */ DefaultNewsService (final @Nonnull Cache cache, final @Nonnull ReadMessagesLog readMessagesLog)
      throws IOException
      {
        this.cache = cache;
        this.readMessagesLog = readMessagesLog;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public synchronized void start()
      {
        log.info("start()");
        
        if (startThread == null)
          {
            if (!networkingPreferences.get().isNetworkConnectionAllowed())
              {
                log.info(">>>> not checking, internet connections are disabled");
              }
            else
              {
                startThread = new Thread()
                  {
                    @Override
                    public void run()
                      {
                        try
                          {
                            updateCache();
                          }
                        catch (Exception e)
                          {
                            log.warn("While running news feed check", e);
                          }

                        startThread = null;
                      }
                  };

                startThread.start();
              }
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public void getNewsFeed (final @Nonnull NewsFeedCallback callBack)
      {
        log.info("getNewsFeed({})", callBack);

        if (cache.getRssFeed() != null) // already loaded in this session
          {
            callBack.notifyFeedAvailable(cache.getRssFeed());
          }
        else
          {
            if (cache.getStatus() == NEEDS_DOWNLOAD) // not initialized?
              {
                try
                  {
                    cache.computeStatus();
                  }
                catch (Exception e)
                  {
                    log.warn("getNewsFeed()", e);
                    callBack.notifyFeedUnavailable();
                    return;
                  }
              }

            if (!networkingPreferences.get().isNetworkConnectionAllowed() &&
                ((cache.getStatus() == NEEDS_DOWNLOAD) || (cache.getStatus() == NEEDS_REFRESH)))
              {
                callBack.notifyFeedCouldBeDownloaded();
              }
            else
              {
                getNewsFeedAlways(callBack);
              }
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public void getNewsFeedAlways (final @Nonnull NewsFeedCallback callBack) 
      {
        new Thread()
          {
            @Override
            public void run()
              {
                log.debug("getNewsFeedAlways({})", callBack);

                try
                  {
                    switch (cache.getStatus())
                      {
                        case DOWNLOADING:
                            waitForDownloaded();
                            break;
                            
                        case NEEDS_DOWNLOAD:
                        case NEEDS_REFRESH:
                            cache.download();
                            break;
                      }

                    readAndCallBack(callBack); 
                  }
                catch (Exception e)
                  {
                    log.warn("getNewsFeedAlways()", e);
                    
                    if (cache.getStatus() == NEEDS_REFRESH)
                      {
                        readAndCallBack(callBack);
                      }
                    else
                      {
                        callBack.notifyFeedUnavailable();
                      }
                  }
              }
          }.start();
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ void readAndCallBack (final @Nonnull NewsFeedCallback callBack)
      {
        try
          {
            final RssFeed rssFeed = cache.read();
            
            if (cache.getStatus() == NEEDS_REFRESH)
              {
                callBack.notifyCachedFeedAvailable(rssFeed);                    
              }
            else
              {
                callBack.notifyFeedAvailable(rssFeed);
              }
          }
        catch (Exception e)
          {
            log.warn("Error calling notifyFeedAvailable() or notifyObsoleteFeedAvailable()", e.toString());
            callBack.notifyFeedUnavailable();
          }
      }

    /*******************************************************************************************************************
     *
     * Called in background to verify whether the cache needs updating. If there are new messages, posts a 
     * notification.
     *
     ******************************************************************************************************************/
    /* package */ void updateCache()
      throws Exception
      {
        log.info("updateCache()");
        cache.computeStatus();

        if (cache.getStatus() == NEEDS_DOWNLOAD)
          {
            cache.download(60000); // running in background, longer timeout
            cache.read();
            eventuallyPostUnreadNewsNotification();
          }
      }

    /*******************************************************************************************************************
     *
     * Checks whether there are unread news and eventually posts a related notification.
     *
     ******************************************************************************************************************/
    /* package */ void eventuallyPostUnreadNewsNotification()
      {
        log.info("eventuallyPostUnreadNewsNotification()");

        if (!readMessagesLog.isAllRead(cache.getRssFeed()))
          {
            postUnreadNewsNotification();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void postUnreadNewsNotification()
      {
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public void markRead (final @Nonnull Document message)
      throws IOException
      {
        readMessagesLog.markRead(message);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public void markNotRead (final @Nonnull Document message)
      throws IOException
      {
        readMessagesLog.markNotRead(message);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public void markAllRead (final @Nonnull RssFeed rssFeed)
      throws IOException
      {
        readMessagesLog.markAllRead(rssFeed);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public boolean isAllRead (final @Nonnull RssFeed rssFeed)
      {
        return readMessagesLog.isAllRead(rssFeed);
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean isRead (final @Nonnull Document message)
      {
        return readMessagesLog.isRead(message);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected URL createUrl (final String string)
      throws IOException
      {
        return new URL(string);
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void waitForDownloaded() 
      throws IOException
      {
        synchronized (cache)
          {
            while (cache.getStatus() == DOWNLOADING)
              {
                log.debug(">>>> status is {}, waiting...", cache.getStatus());
                try
                  {
                   cache.wait();
                  } 
                catch (InterruptedException e) 
                  {
                    // FIXME
                  }
              }
          }
        
        log.debug(">>>> status is {}", cache.getStatus());
        
        if (cache.getStatus() != DOWNLOADED)
          {
            throw new IOException("Download did not complete");  
          }
      }
  }
