/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.news;

import javax.swing.Action;
import javax.swing.AbstractAction;
import java.awt.event.ActionEvent;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.io.IOException;
import it.tidalwave.util.Task;
import it.tidalwave.util.RoleFactory;
import it.tidalwave.role.ui.PresentationModel;
import it.tidalwave.role.ui.spi.DefaultPresentationModel;
import it.tidalwave.semantic.document.Document;
import it.tidalwave.semantic.rss.RssFeed;
import it.tidalwave.mobile.ui.QuestionWithFeedback;
import it.tidalwave.bluebill.mobile.role.util.RoleRegister;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.capabilitiesprovider.ThreadLookupBinder;
import it.tidalwave.bluebill.mobile.news.impl.NewsItemActionProvider;
import it.tidalwave.util.thread.ThreadType;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import static org.openide.util.NbBundle.*;
import static it.tidalwave.semantic.document.Document.Document;

/***********************************************************************************************************************
 *
 * @stereotype Controller
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public abstract class DefaultNewsViewController implements NewsViewController
  {
    private static final Logger log = LoggerFactory.getLogger(DefaultNewsViewController.class);
    
    private static final Class<DefaultNewsViewController> _ = DefaultNewsViewController.class;
    
    private static final Class<RssFeed> RssFeed = RssFeed.class;
    
    @Nonnull
    private final NewsView view;
    
    private final Provider<NewsService> newsService = Locator.createProviderFor(NewsService.class);
    
    protected PresentationModel newsFeedPM;
    
    protected final RoleRegister roleRegister = new RoleRegister();    
    
    /*******************************************************************************************************************
     *
     * An {@code Action} that marks all messages as read.
     *
     ******************************************************************************************************************/
    @Getter @SuppressWarnings("serial")
    private final Action markAllMessagesAsReadAction = new AbstractAction(getMessage(_, "markAllMessagesAsRead")) 
      {
        public void actionPerformed (final @Nonnull ActionEvent event) 
          {
            try 
              {
                log.info("markAllMessagesAsReadAction.actionPerformed()");
                newsService.get().markAllRead(newsFeedPM.as(RssFeed.class));
                view.populate(newsFeedPM); // FIXME: fire an event instead to rerender the single message
                view.notifyAllMessagesMarkedAsRead(getMessage(_, "allMarkedAsRead"));
                setEnabled(false);
              }
            catch (IOException e)
              {
                log.error("setRead()", e);
                // TODO: notify the UI?
              }
          }
      };
    
    /*******************************************************************************************************************
     *
     * The default implementation of {@link Readable} for messages in the news feed.
     *
     ******************************************************************************************************************/
    @RequiredArgsConstructor
    class DefaultReadable implements Readable
      {
        @Nonnull
        private final PresentationModel messagePM;

        public boolean isRead() 
          {
            return newsService.get().isRead(messagePM.as(Document));
          }
        
        public void markAsRead()
          {
            try 
              {
                newsService.get().markRead(messagePM.as(Document));
                markAllMessagesAsReadAction.setEnabled(!newsService.get().isAllRead(newsFeedPM.as(RssFeed)));
              }
            catch (IOException e)
              {
                log.error("setRead()", e);
                // TODO: notify the UI?
              }
            
            view.populate(newsFeedPM); // FIXME: fire an event instead to rerender the single message
          }
        
        public void markAsNotRead()
          {
            try 
              {
                newsService.get().markNotRead(messagePM.as(Document));
                markAllMessagesAsReadAction.setEnabled(true);
              }
            catch (IOException e)
              {
                log.error("setRead()", e);
                // TODO: notify the UI?
              }
            
            view.populate(newsFeedPM); // FIXME: fire an event instead to rerender the single message
          }
        
        public void read()
          {
            readMessage(messagePM.as(Document));
            markAsRead();
          }
      }
    
    /*******************************************************************************************************************
     *
     * FIXME: merge this and the one below in a CapabilityProvider.
     *
     ******************************************************************************************************************/
    private final RoleFactory<PresentationModel> messageActionProviderFactory = new RoleFactory<PresentationModel>() 
      {
        @Nonnull
        public Object createRoleFor (final @Nonnull PresentationModel messagePM)  
          {
            return new NewsItemActionProvider(messagePM);
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final RoleFactory<PresentationModel> readableFactory = new RoleFactory<PresentationModel>() 
      {
        @Nonnull
        public Object createRoleFor (final @Nonnull PresentationModel messagePM)  
          {
            return new DefaultReadable(messagePM);
          }
      };

    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    private final NewsService.NewsFeedCallback newsServiceCallback = new NewsService.NewsFeedCallback() 
      {
        @ThreadConfined(type=ThreadType.NOT_UI)
        public void notifyFeedAvailable (final @Nonnull RssFeed newsFeed) 
          {
            log.debug("notifyFeedAvailable()");
            populateView(newsFeed);
          }

        @ThreadConfined(type=ThreadType.NOT_UI)
        public void notifyCachedFeedAvailable (final @Nonnull RssFeed newsFeed) 
          {
            log.debug("notifyCachedFeedAvailable()");
            populateView(newsFeed);
            view.notifyFeedIsCached(getMessage(_, "obsoleteNews"));
          }

        @ThreadConfined(type=ThreadType.NOT_UI)
        public void notifyFeedUnavailable() 
          {
            log.debug("notifyFeedUnavailable()");
            markAllMessagesAsReadAction.setEnabled(false);
            view.dismissPleaseWaitAndUnlockUI();
            view.notifyFeedUnavailable(new QuestionWithFeedback(getMessage(_, "unavailableNewsTitle"), 
                                                                getMessage(_, "unavailableNewsMessage"))
              {
                @Override
                public void onConfirm() 
                  {
                    view.dismiss();
                  }
              });
          }

        @ThreadConfined(type=ThreadType.NOT_UI)
        public void notifyFeedCouldBeDownloaded() 
          {
            log.debug("notifyFeedCouldBeDownloaded()");
            view.confirmToDownloadNews(new QuestionWithFeedback(getMessage(_, "confirmDownloadTitle"),
                                                                getMessage(_, "confirmDownloadMessage"))
              {
                @Override
                public void onConfirm()
                  {
                    newsService.get().getNewsFeedAlways(newsServiceCallback);
                  }

                @Override
                public void onCancel()
                  {
                    view.dismissPleaseWaitAndUnlockUI();
                    view.dismiss();
                  }
              });
          }
      };
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public DefaultNewsViewController (final @Nonnull NewsView view) 
      {
        this.view = view;
        roleRegister.registerRoleFactory(messageActionProviderFactory).forClass(DefaultPresentationModel.class);
        roleRegister.registerRoleFactory(readableFactory).forClass(DefaultPresentationModel.class);
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public void loadNewsFeed()
      {
        log.info("loadNewsFeed()");
        view.showPleaseWaitAndLockUI(getMessage(_, "preparingNews")); 
        
        // TODO: add a timeout and eventually show an error
        newsService.get().getNewsFeed(newsServiceCallback);
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void cancelLoading() 
      {
        log.info("progressDialogCanceled()");
        view.dismissPleaseWaitAndUnlockUI();
        view.dismiss();
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ThreadType.NOT_UI)
    private void populateView (final @Nonnull RssFeed newsFeed)
      {
        ThreadLookupBinder.with(roleRegister).run(new Task<Void, RuntimeException>() 
          {
            @Override
            public Void run() 
              {
                log.debug("populateView({})", newsFeed);
                newsFeedPM = new DefaultPresentationModel(newsFeed);
                markAllMessagesAsReadAction.setEnabled(!newsService.get().isAllRead(newsFeedPM.as(RssFeed)));                        
                view.populate(newsFeedPM);
                view.dismissPleaseWaitAndUnlockUI();
                return null;
              }
          });
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected abstract void readMessage (@Nonnull Document message);
    
    @Nonnull
    /* package */ Readable createReadable (final @Nonnull PresentationModel messagePM)
      {
        return new DefaultReadable(messagePM);  
      }
  }
