/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.news.impl;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.util.Date;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.mobile.io.FileSystem;
import it.tidalwave.mobile.io.IoUtils;
import it.tidalwave.mobile.io.MasterFileSystem;
import it.tidalwave.semantic.rss.RssFeed;
import it.tidalwave.semantic.rss.RssFeedReader;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import lombok.Cleanup;
import lombok.Getter;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class Cache 
  {
    public enum Status
      {
        NEEDS_DOWNLOAD,
        NEEDS_REFRESH,
        DOWNLOADING,
        DOWNLOADED
      }

    private static final Logger log = LoggerFactory.getLogger(Cache.class);

    public static final long DOWNLOAD_TIMEOUT = 15000;
    
    private static final long ONE_DAY = 24 * 60 * 60 * 1000;
    private static final long ONE_MINUTE = 60 * 1000;

    @CheckForNull @Getter
    private RssFeed rssFeed; 

    @Nonnull @Getter
    /* package */ volatile Status status = Status.NEEDS_DOWNLOAD; 
    
    @Nonnegative
    /* package */ long refreshPeriod = ONE_DAY; 

    @Nonnull
    private final URL url;
    
    @Nonnull
    /* package */ final File cachedFile; 

    @Nonnull
    private final Provider<MasterFileSystem> masterFileSystem = Locator.createProviderFor(MasterFileSystem.class);

    /*******************************************************************************************************************
     *
     * FIXME: replace with Downloadable, so you can use rssFeed.as(Downloadable).
     *
     ******************************************************************************************************************/
    public Cache (final @Nonnull URL url) 
      throws IOException
      {
        final FileSystem internalFileSystem = masterFileSystem.get().getInternalFileSystem();
        cachedFile = internalFileSystem.getFile("blog.rss");
        this.url = url;
      }
    
    /*******************************************************************************************************************
     *
     * Downloads the news in the current thread, blocking until they are ready.
     *
     ******************************************************************************************************************/
    public void download()
      throws IOException
      {
        download(DOWNLOAD_TIMEOUT);
      }
    
    /*******************************************************************************************************************
     *
     * Downloads the news in the current thread, blocking until they are ready.
     *
     ******************************************************************************************************************/
    public void download (final @Nonnegative long timeout)
      throws IOException
      {
        log.info("download({} msec)", timeout);
        final Status previousStatus = getStatus();
        setStatus(Status.DOWNLOADING);
        // TODO: be smart, check the last update time of the feed
        final AtomicBoolean canceled = new AtomicBoolean(false);
        final AtomicBoolean done = new AtomicBoolean(false);
        final AtomicReference<Exception> exception = new AtomicReference<Exception>();
        
        final Thread thread = new Thread()
          {
            @Override
            public void run()
              {
                try
                  {
                    final File downloadFile = doDownload();
                    
                    if (!canceled.get())
                      {
                        cachedFile.delete();
                        downloadFile.renameTo(cachedFile);
                        done.set(true);
                      }
                  }
                catch (IOException e)
                  {
                    log.warn("download()", e);
                    exception.set(e);
                  }
              }
          };
        
        thread.start();
        
        try 
          {
            thread.join(timeout);
          }
        catch (InterruptedException e) 
          {
            // FIXME
          }
        
        if (exception.get() != null)
          {
            setStatus(previousStatus);
            throw new IOException(exception.get().toString()); // Java 5 / Android compatibility  
          }
        else if (done.get())
          {
            setStatus(Status.DOWNLOADED);
          }
        else // timed out
          {
            log.warn(">>>> timed out");
            canceled.set(true);
            thread.interrupt();
            setStatus(previousStatus);
            throw new IOException("Timed out");
          }
      }
    
    /*******************************************************************************************************************
     *
     * In a separate method so it can be mocked in tests.
     *
     ******************************************************************************************************************/
    @Nonnull
    /* package */ File doDownload()
      throws IOException
      {
        final File tempFile = new File(cachedFile.getAbsolutePath() + ".download");
        @Cleanup final InputStream is = url.openStream();
        @Cleanup final OutputStream os = new FileOutputStream(tempFile);
        IoUtils.copy(is, os);
        return tempFile;
      }
    
    /*******************************************************************************************************************
     *
     * Loads the news feed from the cache.
     *
     ******************************************************************************************************************/
    @Nonnull
    public RssFeed read() 
      throws IOException
      {
        try
          {
            @Cleanup final InputStream is = new FileInputStream(cachedFile);
            final RssFeedReader reader = new RssFeedReader();
            return rssFeed = reader.readMessages(is);
          }
        catch (Exception e)
          {
            cachedFile.delete();
            rssFeed = null;
            setStatus(Status.NEEDS_DOWNLOAD);
            throw new IOException("Error reading feed: " + e.toString()); // Java 5 / Android compatibility
          }
      }
    
    /*******************************************************************************************************************
     *
     * COmputes the cache status. The status is updated as:
     * 
     * <ul>
     * <li>{@code NEEDS_DOWNLOAD} if the cache is empty;</li>
     * <li>{@code NEEDS_REFRESH} if the cache is outdated;</li>
     * <li>{@code DOWNLOADED} if the cache is not empty and updated.</li>
     * </ul>
     *
     ******************************************************************************************************************/
    public void computeStatus()
      throws Exception
      {
        log.info("computeStatus()");
        log.info(">>>> cached file is {}", cachedFile.getAbsolutePath());

        if (!cachedFile.exists())
          {
            log.info(">>>> cached file doesn't exist, needs download");
            setStatus(Status.NEEDS_DOWNLOAD);
          }
        else
          {
            final long lastModified = cachedFile.lastModified();
            final long now = getCurrentTime();
            final long delta = now - lastModified;

            setStatus((delta >= refreshPeriod) ? Status.NEEDS_REFRESH : Status.DOWNLOADED);

            log.info(">>>> now: {}, latest news download: {}, delta: {} min, status: {}",
                       new Object[] { new Date(now), new Date(lastModified), delta / ONE_MINUTE, status } );
          }
      }
    
    /*******************************************************************************************************************
     *
     * In a specific method so it can be mocked.
     * 
     * TODO: use a TimeProvider
     *
     ******************************************************************************************************************/
    /* package */ long getCurrentTime() 
      {
        return System.currentTimeMillis();
      }
    
   /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ synchronized void setStatus (final @Nonnull Status status)
      {
        log.info("setStatus({})", status);
        this.status = status;
        notifyAll(); 
      }
   }
